/' -- DEBUGGING -- '/
#define debug
#ifdef debug
sub debugc() constructor
  open cons for output as #-1
  print #-1, "-- DEBUG MODE --"
end sub
sub debugd() destructor
  print #-1, "APP TERMINATE"
  close #-1
  screen -1
  sleep 750, 1
end sub
#endif
/' -- DEBUGGING -- '/

#define fb_bswap(var) cushort((((var) and &hff) shl 8) or (((var) and &hff00) shr 8))

const as ubyte tilew = 24, tileh = 28
dim shared as ushort layerw(2) = { 14,  14,  15}
dim shared as ushort layerh(2) = {300, 600, 600}

type tyr_map_t
  as ushort ptr index(2)
  as ubyte ptr layer(2)
end type

type tyr_shapes_t
  as uinteger ptr tile(600 - 1)
end type

declare function load_palette() as uinteger ptr
declare function load_map(mapfile as zstring ptr) as tyr_map_t ptr
declare function save_map(mapfile as zstring ptr, tyr_map as tyr_map_t ptr) as ubyte
declare function load_shapes(shapesfile as zstring ptr) as tyr_shapes_t ptr
declare function unload_shapes(tyr_shapes as tyr_shapes_t ptr) as ubyte
declare sub draw_index_tiles(tiles_buffer as uinteger ptr, tyr_shapes as tyr_shapes_t ptr, tyr_map as tyr_map_t ptr, layer as ubyte)
declare sub draw_map(layer_buffer as uinteger ptr, tyr_shapes as tyr_shapes_t ptr, tyr_map as tyr_map_t ptr, layer as ubyte, yoff as integer)
declare function fail(as zstring ptr) as ubyte

dim as tyr_map_t ptr tyr_map
dim as tyr_shapes_t ptr tyr_shapes

dim shared as uinteger ptr tyr_pal

/'
'' temporary(?) interface
'/

if command$ = "" then
  print
  print "TyrMap v0.0.2 alpha  [Mindless]"
  print
  print command$(0) & " [file.map] [shapes?.dat]"
  end
end if

screenres 640, 480, 24, 2

tyr_map = load_map(command$(1))
tyr_pal = load_palette()
tyr_shapes = load_shapes(command$(2))

/'
'' end temporary interface
'/

dim as ushort x, y
dim as integer mousex, mousey, mousew, mouseb
dim as integer mousex_delta, mousey_delta, mousew_delta

dim as integer yoff
dim as ubyte redraw_map = 1, redraw_tiles = 1, redraw_hover = 1

dim as integer selected_tile = 0, hover_tile = -1
dim as integer hover_x = -1, hover_y = -1

dim as ubyte layer = 0

const as ushort index_xoff = 640 - tilew * 12

/'
'' graphical buffers
'/

dim as uinteger ptr layer_buffer = imagecreate(15 * tilew, 16 * tileh, rgba(0, 0, 0, 0))
dim as uinteger ptr tiles_buffer = imagecreate(12 * tilew, 11 * tileh, rgba(0, 0, 0, 0))
dim as uinteger ptr hover_buffer = imagecreate(tilew, tileh, rgba(0, 0, 0, 0))
dim as uinteger ptr selected_buffer = imagecreate(tilew, tileh, rgba(0, 0, 0, 0))

'' draw hover and selection boxes
line hover_buffer, (0, 0)-(tilew - 1, tileh - 1), rgba(255, 255, 255, 255), B, &hcccccccc
line selected_buffer, (0, 0)-(tilew - 1, tileh - 1), rgba(255, 255, 255, 200), B, &h66666666
line hover_buffer, (0, 0)-(tilew - 1, tileh - 1), rgba(0, 0, 0, 255), B, &h33333333
line selected_buffer, (0, 0)-(tilew - 1, tileh - 1), rgba(0, 0, 0, 200), B, &h99999999

'' main loop until "esc" key pressed
do until multikey(1)
  
  if multikey(&h48) then yoff -= 1 : redraw_map = 1
  if multikey(&h50) then yoff += 1 : redraw_map = 1
  
  '' check if keys "1", "2" or "3" are being pressed
  dim as ubyte i
  for i = 0 to 2
    if multikey(i + 2) then
      layer = i
      redraw_tiles = 1
      do while multikey(i + 2) : loop
      #ifdef debug
      print #-1, "SWITCH TO LAYER " & i
      #endif
    end if
  next
  
  '' check if "s" key is pressed
  if multikey(&h1f) then save_map(command$(1) & ".revised", tyr_map) : do while multikey(&h1f) : loop
  
  mousex_delta = mousex
  mousey_delta = mousey
  mousew_delta = mousew
  if getmouse(mousex, mousey, mousew, mouseb) = 0 then
    mousex_delta -= mousex
    mousey_delta -= mousey
    mousew_delta -= mousew
  end if
  
  if yoff < 0 then yoff = 0
  if yoff >= layerh(layer) - 16 then yoff = layerh(layer) - 16
  
  screenset 1, 0
  
  if redraw_map or redraw_tiles then
    cls
    
    draw_map(layer_buffer, tyr_shapes, tyr_map, layer, yoff)
    put (0, 0), layer_buffer, pset
    
    if redraw_tiles then
      draw_index_tiles(tiles_buffer, tyr_shapes, tyr_map, layer)
      redraw_tiles = 0
    end if
    put (640 - tilew * 12, 0), tiles_buffer, pset
    
    redraw_map = 0
  end if
  
  screenlock
  screencopy
  screenset 0, 0
  
  if mousex >= 0 and mousex <= layerw(layer) * tilew - 1 and mousex < index_xoff then
    if mousey >= 0 and mousey <= 16 * tileh - 1 then
      hover_y = mousey \ tileh
      hover_x = mousex \ tilew
      put (hover_x * tilew, hover_y * tileh), hover_buffer, alpha
      if mouseb and 1 then
        tyr_map->layer(layer)[(hover_y + yoff) * layerw(layer) + hover_x] = selected_tile
        redraw_map = 1
      end if
      if mouseb and 2 then
        selected_tile = tyr_map->layer(layer)[(hover_y + yoff) * layerw(layer) + hover_x]
      end if
      draw string (0, 480 - 8), hover_x & ", " & hover_y + yoff
    end if
  end if
  
  if mousex >= index_xoff and mousex <= (640 - tilew * 12) + 15 * tilew - 1 then
    if mousey >= 0 and mousey <= 11 * tileh - 1 then
      hover_y = mousey \ tileh
      hover_x = (mousex - index_xoff) \ tilew
      put (index_xoff + hover_x * tilew, hover_y * tileh), hover_buffer, alpha
      if mouseb and 1 then
        selected_tile = (hover_y) * 12 + (hover_x)
        if selected_tile > 127 then selected_tile = 127
      end if
    end if
  end if
  
  put (index_xoff + (selected_tile mod 12) * tilew, (selected_tile \ 12) * tileh), selected_buffer, alpha
  
  screenunlock
  
  sleep 1, 1
loop

unload_shapes(tyr_shapes)

imagedestroy(layer_buffer)
imagedestroy(tiles_buffer)
imagedestroy(hover_buffer)
imagedestroy(selected_buffer)



function load_palette() as uinteger ptr
  dim as uinteger ptr pal = allocate(256 * sizeof(uinteger))
  
  dim as ubyte fh = freefile
  if open("palette.dat" for binary access read as #fh) then fail(@__function__)
  seek #fh, 1280 * 3 + 1
  dim as ushort i
  dim as ubyte r, g, b
  for i = 0 to 255
    get #fh, , r, 1
    get #fh, , g, 1
    get #fh, , b, 1
    pal[i] = rgb(r * 4, g * 4, b * 4)
  next
  close #fh
  
  return pal
end function

function load_shapes(shapesfile as zstring ptr) as tyr_shapes_t ptr
  dim as tyr_shapes_t ptr tyr_shapes = allocate(sizeof(tyr_shapes_t))
  dim as ushort i, x, y
  dim as ubyte temp
  dim as ubyte ptr tempdata = allocate(24 * 28)
  
  for i = 0 to 600 - 1
    tyr_shapes->tile(i) = imagecreate(24, 28, rgba(0, 0, 0, 0))
  next
  
  dim as ushort fh = freefile
  if open(*shapesfile for binary access read as #fh) then print fail(@__function__)
  
  i = 0
  do until eof(fh) or i > 600 - 1
    get #fh, , temp
    if temp = 0 then
      get #fh, , *tempdata, 24 * 28
      for x = 0 to 24 * 28 - 1
        tyr_shapes->tile(i)[x + 8] = tyr_pal[tempdata[x]] or iif(tempdata[x] = 0, 0, 255 shl 24)
      next
      i += 1
    else
      i += temp
    end if
  loop
  
  close #fh
  
  deallocate(tempdata)
  
  return tyr_shapes
end function

function unload_shapes(tyr_shapes as tyr_shapes_t ptr) as ubyte
  dim as ushort i
  for i = 0 to 600 - 1
    imagedestroy(tyr_shapes->tile(i))
  next
  
  deallocate(tyr_shapes)
  
  return 0
end function

function load_map(mapfile as zstring ptr) as tyr_map_t ptr
  dim as tyr_map_t ptr tyr_map = allocate(sizeof(tyr_map_t))
  
  dim as ushort i, j
  
  dim as ushort fh = freefile
  if open(*mapfile for binary access read as #fh) then fail(@__function__)
  
  for j = 0 to 2
    tyr_map->index(j) = allocate(128 * sizeof(ushort))
    
    for i = 0 to 128 - 1
      get #fh, , tyr_map->index(j)[i], 1
      tyr_map->index(j)[i] = fb_bswap(tyr_map->index(j)[i])
    next
  next
  
  for i = 0 to 2
    tyr_map->layer(i) = allocate(layerh(i) * layerw(i))
    get #fh, , *tyr_map->layer(i), layerh(i) * layerw(i)
  next
  
  close #fh
  
  return tyr_map
end function

function save_map(mapfile as zstring ptr, tyr_map as tyr_map_t ptr) as ubyte
  dim as ushort i, j
  
  dim as ushort fh = freefile
  if open(*mapfile for binary access write as #fh) then fail(@__function__)
  
  for j = 0 to 2
    for i = 0 to 128 - 1
      put #fh, , fb_bswap(tyr_map->index(j)[i]), 1
    next
  next
  
  for i = 0 to 2
    put #fh, , *tyr_map->layer(i), layerh(i) * layerw(i)
  next
  
  close #fh
  
  #ifdef debug
  print #-1, "SAVED AS " & *mapfile & ".revised"
  #endif
  
  return 0
end function

sub draw_index_tiles(tiles_buffer as uinteger ptr, tyr_shapes as tyr_shapes_t ptr, tyr_map as tyr_map_t ptr, layer as ubyte)
  dim as ushort x, y
  for y = 0 to 10
    for x = 0 to iif(y < 10, 11, 7)
      if tyr_map->index(layer)[y * 12 + x] > 0 then
        put tiles_buffer, (x * tilew, y * tileh), tyr_shapes->tile(tyr_map->index(layer)[y * 12 + x] - 1), pset', alpha
      end if
    next
  next
  
  #ifdef debug
  print #-1, "TILES DRAWN"
  #endif
  
end sub

sub draw_map(layer_buffer as uinteger ptr, tyr_shapes as tyr_shapes_t ptr, tyr_map as tyr_map_t ptr, layer as ubyte, yoff as integer)
  dim as ushort x, y
  line layer_buffer, (0, 0)-(15 * tilew - 1, 16 * tileh - 1), &h00000000, BF
  
  for y = 0 to 16 - 1
    for x = 0 to layerw(layer) - 1
      put layer_buffer, (x * tilew, y * tileh), tyr_shapes->tile(tyr_map->index(layer)[tyr_map->layer(layer)[(y + yoff) * layerw(layer) + x]] - 1), pset', alpha
    next
  next
  
  #ifdef debug
  print #-1, "MAP DRAWN"
  #endif
  
end sub


function fail(e as zstring ptr) as ubyte
  print "failure in " & *e
  sleep
  end
  return 0
end function
